#include "Main.h"
#include "Network.h"

#include <chrono>
#include <iostream>
#include <fstream>
#include <SDL.h>
#include <random>
#include <Windows.h>

int reverseInt(int i)
{
	unsigned char c1, c2, c3, c4;

	c1 = i & 255;
	c2 = (i >> 8) & 255;
	c3 = (i >> 16) & 255;
	c4 = (i >> 24) & 255;

	return ((int)c1 << 24) + ((int)c2 << 16) + ((int)c3 << 8) + c4;
}

int OpenFile(std::ifstream &file, Type type)
{
	std::string filepath = "Files/";

	switch (type)
	{
	case Type::TrainLabels:
		filepath += "train-labels.idx1-ubyte";
		break;
	case Type::TrainImages:
		filepath += "train-images.idx3-ubyte";
		break;
	case Type::TestLabels:
		filepath += "t10k-labels.idx1-ubyte";
		break;
	case Type::TestImages:
		filepath += "t10k-images.idx3-ubyte";
		break;
	}

	file.open(filepath, std::ios::in | std::ios::binary);
	if (!file)
	{
		printf("File not found");
		abort();
	}

	int MagicNumber = 0;
	file.read((char*)&MagicNumber, sizeof(int));
	MagicNumber = reverseInt(MagicNumber);

	if (type == Type::TrainLabels || type == Type::TestLabels)
	{
		if (MagicNumber != 2049)
		{
			printf("Magic number not correct");
			abort();
		}

		int Items = 0;
		file.read((char*)&Items, sizeof(int));
		Items = reverseInt(Items);

		printf("Loaded %i Labels\n", Items);
		return Items;
	}
	else
	{
		if (MagicNumber != 2051)
		{
			printf("Magic number not correct");
			abort();
		}

		int Images = 0;
		int Rows = 0;
		int Columns = 0;
		file.read((char*)&Images, sizeof(int));
		file.read((char*)&Rows, sizeof(int));
		file.read((char*)&Columns, sizeof(int));
		Images = reverseInt(Images);
		Rows = reverseInt(Rows);
		Columns = reverseInt(Columns);

		if (Rows != 28 || Columns != 28)
		{
			printf("Size is not correct");
			abort();
		}

		printf("Loaded %i Images\n", Images);
		return Images;
	}
}

void DrawNumber(SDL_Renderer* Renderer, int Number, int XPos, int YPos)
{
	std::string pixel;

	switch (Number)
	{
	case 0: pixel = ":ooo:  o:::o  o::oo  o:o:o  oo::o  o:::o  :ooo:"; break;
	case 1: pixel = "::o::  :oo::  ::o::  ::o::  ::o::  ::o::  :ooo:"; break;
	case 2: pixel = ":ooo:  o:::o  ::::o  :::o:  ::o::  :o:::  ooooo"; break;
	case 3: pixel = "ooooo  :::o:  ::o::  :::o:  ::::o  o:::o  :ooo:"; break;
	case 4: pixel = ":::o:  ::oo:  :o:o:  o::o:  ooooo  :::o:  :::o:"; break;
	case 5: pixel = "ooooo  o::::  oooo:  ::::o  ::::o  o:::o  :ooo:"; break;
	case 6: pixel = "::oo:  :o:::  o::::  oooo:  o:::o  o:::o  :ooo:"; break;
	case 7: pixel = "ooooo  ::::o  :::o:  ::o::  :o:::  :o:::  :o:::"; break;
	case 8: pixel = ":ooo:  o:::o  o:::o  :ooo:  o:::o  o:::o  :ooo:"; break;
	case 9: pixel = ":ooo:  o:::o  o:::o  :oooo  ::::o  :::o:  :oo::"; break;
	default: return;
	}

	for (int i = 0; i < 47; i++)
	{
		if (pixel[i] != ':' && pixel[i] != ' ')
		{
			SDL_RenderDrawPoint(Renderer, XPos + i % 7, YPos + (int)floor((float)i / 7.f));
		}
	}
}

int main(int argc, char** argv)
{
	NeuralNetwork nn(784, 30, 10);

	std::ifstream labels;
	std::ifstream images;
	int TotalItems;
	int TotalImages;

	printf( "Press numbers 1-8 to change the speed\n"
			"Press number 9 to set the speed to unlimited\n"
			"Press number 0 to pause\n\n");

	TotalItems = OpenFile(labels, Type::TrainLabels);
	TotalImages = OpenFile(images, Type::TrainImages);
	if (TotalItems != TotalImages)
		abort();

	printf("\nStarted training\n");

	bool ExitApp = false;
	SDL_Event Event;
	int FrameCount = 0;
	float delta_time = 0.0001f;
	bool EndOfFile = false;
	bool IsTesting = false;
	int Speed = 1;
	int GuesssedCorrect = 0;
	int GuesssedIncorrect = 0;
	bool Finished = false;

	SDL_Init(SDL_INIT_VIDEO);

	SDL_Window* screen = SDL_CreateWindow("Sandbox", SDL_WINDOWPOS_CENTERED_DISPLAY(DISPLAY), SDL_WINDOWPOS_CENTERED_DISPLAY(DISPLAY), SCRWIDTH, SCRHEIGHT, SDL_WINDOW_BORDERLESS);
	SDL_Renderer* renderer = SDL_CreateRenderer(screen, -1, 0);
	SDL_SetRenderDrawColor(renderer, 40, 40, 60, 0);
	SDL_RenderClear(renderer);

	while (!ExitApp)
	{
		auto start = std::chrono::steady_clock::now();

		// if we've paused we don't want to teleport everything ahead by a lot
		if (delta_time > 0.1f)
			delta_time = 0.1f;

		SDL_PollEvent(&Event);				

		if (Speed != 0 && !Finished)
		{
			// prints
			if (FrameCount % PRINTPERCENTAGE == 0)
			{
				if (!IsTesting && FrameCount <= 60000)
					printf("Training: %.1f%%\n", (float)FrameCount / (float)TotalItems * 100.f);
				else if (IsTesting && FrameCount <= 10000)
				{
					printf("Testing: %.1f%%, Correct: %.1f%%\n",
						(float)(FrameCount) / (float)TotalItems * 100.f,
						((float)GuesssedCorrect / (float)(GuesssedCorrect + GuesssedIncorrect)) * 100);
				}
			}

			bool ShouldRender = Speed > 500 ? FrameCount % PRINTPERCENTAGE == 0 : true;

			// rendering
			if (!EndOfFile)
			{
				// read label
				unsigned char label = 0;
				labels.read((char*)&label, sizeof(unsigned char));
				int target = (int)label;
				if (labels.eof())
				{
					EndOfFile = true;
				}
				else
				{
					// draw image
					std::vector<double> inputs(784);
					unsigned char pixel = 0;
					int value = 0;
					if (ShouldRender)
						SDL_RenderSetScale(renderer, IMAGESIZE, IMAGESIZE);

					for (int y = 0; y < 28; y++)
					{
						for (int x = 0; x < 28; x++)
						{
							images.read((char*)&pixel, sizeof(unsigned char));
							value = (int)pixel;
							inputs[x * y] = (pixel / 255.f);

							if (ShouldRender)
							{
								SDL_SetRenderDrawColor(renderer, value, value, value, 0);
								SDL_RenderDrawPoint(renderer, 10 + x, 10 + y);
							}
						}
					}

					// do actual neural network
					nn.feedForward(inputs);
					std::vector<double> outputs = nn.getOutput();

					// get guessed number
					float HighestValue = 0.f;
					int GuessedNumber = -1;
					for (int i = 0; i < 10; i++)
					{
						if (outputs[i] >= HighestValue)
						{
							HighestValue = static_cast<float>(outputs[i]);
							GuessedNumber = i;
						}
					}

					if (!IsTesting)
					{
						// train the network
						std::vector<double> temp(10);
						for (int i = 0; i < 10; i++)
						{
							temp[i] = (i == target) ? 1 : 0;
						}
						nn.backPropagate(temp);
					}
					else
					{
						if (target == GuessedNumber)
							GuesssedCorrect++;
						else
							GuesssedIncorrect++;
					}

					if (ShouldRender)
					{
						// draw numbers
						SDL_RenderSetScale(renderer, TEXTSIZE, TEXTSIZE);
						for (int i = 0; i < 10; i++)
						{
							int color = (int)round(outputs[i] * 255.f);
							SDL_SetRenderDrawColor(renderer, color, color, color, 0);
							DrawNumber(renderer, i, 110, 10 + (TEXTSIZE - 1) * i);
						}

						// draw final answer
						SDL_RenderSetScale(renderer, IMAGESIZE, IMAGESIZE);

						if (target == GuessedNumber)
							SDL_SetRenderDrawColor(renderer, 0, 255, 0, 0);
						else
							SDL_SetRenderDrawColor(renderer, 255, 0, 0, 0);

						DrawNumber(renderer, GuessedNumber, 60, 20);

						// present the renderer to the window
						SDL_RenderPresent(renderer);
						SDL_SetRenderDrawColor(renderer, 40, 40, 60, 0);
						SDL_RenderClear(renderer);
					}
				}
			}
			else if (!IsTesting)
			{
				// switch from training to testing
				// stop training
				printf("Training completed\n\n");
				labels.close();
				images.close();
				EndOfFile = false;
				FrameCount = 0;

				// start testing
				IsTesting = true;
				TotalItems = OpenFile(labels, Type::TestLabels);
				TotalImages = OpenFile(images, Type::TestImages);
				if (TotalItems != TotalImages)
					abort();
				printf("Started testing\n");
			}
			else
			{
				printf("Testing completed\n\n");
				printf("Final Correct: %.1f%%\n",
					((float)GuesssedCorrect / (float)(GuesssedCorrect + GuesssedIncorrect)) * 100);
				printf("Press Esc to close");
				Finished = true;
			}
		}

		switch (Event.type)
		{
		case SDL_QUIT:
			ExitApp = true;
			break;
		case SDL_KEYDOWN:
			if (Event.key.keysym.sym == SDLK_ESCAPE) {
				ExitApp = true;
			}
			else if (!Finished)
			{
				bool ChangedSpeed = false;
				switch (Event.key.keysym.sym)
				{
				case SDLK_1: Speed = 1 << 0; ChangedSpeed = true; break;
				case SDLK_2: Speed = 1 << 1; ChangedSpeed = true; break;
				case SDLK_3: Speed = 1 << 2; ChangedSpeed = true; break;
				case SDLK_4: Speed = 1 << 3; ChangedSpeed = true; break;
				case SDLK_5: Speed = 1 << 4; ChangedSpeed = true; break;
				case SDLK_6: Speed = 1 << 5; ChangedSpeed = true; break;
				case SDLK_7: Speed = 1 << 6; ChangedSpeed = true; break;
				case SDLK_8: Speed = 1 << 7; ChangedSpeed = true; break;
				case SDLK_9: Speed = 100000; ChangedSpeed = true; break;
				case SDLK_0: Speed = 0; ChangedSpeed = true; break;
				}

				if (ChangedSpeed)
				{
					if (Speed != 0 && Speed < 500)
						printf("Speed = %i\n", Speed);
					else if (Speed > 500)
						printf("Speed = Unlimited\n");
					else
						printf("Speed = Paused\n");
				}
			}
		default:
			break;
		}

		if (Speed != 0 && !Finished)
			FrameCount++;

		// lock fps
		if (FPSCAP != -1)
		{
			auto end = std::chrono::steady_clock::now();
			std::chrono::duration<float> elapsed_chrono_time = end - start;
			delta_time = elapsed_chrono_time.count();

			float time = 1.f / (float)(FPSCAP * Speed);
			if (delta_time < time)
			{
				Sleep(DWORD((time - delta_time) * 1000.f));
				delta_time = time;
			}
		}
		else
		{
			auto end = std::chrono::steady_clock::now();
			std::chrono::duration<float> elapsed_chrono_time = end - start;
			delta_time = elapsed_chrono_time.count();
		}
	}

	// Quit game
	SDL_Quit();

	labels.close();
	images.close();

	return 0;
}